unit pomocne;

{***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************}
 
{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Dialogs, Math, {libc,}
  {$IFDEF UNIX}
  oldlinux,
  {$ENDIF}
  svd, ap, blas;

//const

  //vel_prior = 2;	{default value of velocities (a priori information)}
  //vel_uncert = 0.1;	{standard deviation of initial value of velocities}
  //depth_prior = 2;	{default (initial) value of depths (a priori information}
  //depth_uncert = 1;	{standard deviation of initial value of depths}
  //data_uncert = 0.1;	{standard deviation of travel time errors}
  //min_vel = 0.5;	{minimal velocity allowed in model}
  //max_vel = 6;		{maximal velocity allowed in model}
  //vel_std = 100;        {the standard deviation which will be assigned to a velocity
                        //value out of range [min_vel; max_vel]. This velocity will
                        //be deleted and the vel_prior value with vel_std will be
                        //assigned instead}
  //min_depth = 0.2;	{minimal depth allowed in model}
  //max_depth = 5;	{maximal depth allowed in model}
  //tol = 0.001;		{the smallest singular value allowed in SVD}
  //err = 0.3;

  //NumZero = 0.000000001; {Cisla mensi nez NumZero se budou povazovat za nulu}

  //MinTime = 0.001; {the smallest arrival time which will be considered}


type
  TTimeData = array of array[0..4] of Double;  {cislo geofonu -- cislo
    odpalu [0], cas prvniho nasazeni [1], cislo vrstvy [2], identifikator
    geofonu (ukazatel na geofon do pole souradnic) [3] a pozice geofonu v matici
    soustavy A [4] -- zaznamy se stejnym identifikatorem jsou ze stejneho geofonu}
  T3dDoubleArray = array of array of array of double;
  T2DDoubleArray = array of array of double;
  T2dIntegerArray = array of array of integer;
  TCoord = array of array[0..1] of double; {souradnice bodu odpalu -- cislo
    odpalu, x [0] a y [1] souradnice}
  TCtverCoord = array of array[0..3] of double; {souradnice x [0], y [1] bunky,
    v [2] je vzdalenost, kterou paprsek tou kterou bunkou urazi, v [3] je index
    bunky (ctverecku)}
  TDoubleArray = array of double;
  TOdpGf = array of array[0..2] of integer; {v prvku [0] je cislo odpalu ktery
    je na stejnem miste jako geofon s identifikatorem v [1] a poctem vrstev na
    danem miste v [2]}

procedure SpoctiVp(const TimeData: TTimeData; var Rychlosti: T3dDoubleArray;
  var VpZname: T2dIntegerArray; const ShotCoord, GeophCoord: TCoord);
procedure InterpolujVp(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double);
{interpoluje rychlosti v zadane vrstve ( 0 = 1. vrstva, 1 = 2. vrstva atd.).
Pokud je bunka [x,y] ve VpZname = 0, neni v tomto bode znama Vp. Vp se spocte
jako aritmeticky prumer rychlost ctyr okolnich bunek. Probiha to iteracemi,
pocet iteraci urcuje iter.}
procedure InterpolujV0(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double);
{interpoluje rychlosti v zadane vrstve ( 0 = 1. vrstva, 1 = 2. vrstva atd.).
Vp se spocte jako aritmeticky prumer rychlost ctyr okolnich bunek. Probiha to
iteracemi, pocet iteraci urcuje iter.}
procedure InterpolujVpXProfil(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double); {Totez co InterpolujVp jenom
  pro profilova mereni ve smeru X}
procedure InterpolujVpYProfil(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double); {Totez co InterpolujVp jenom
  pro profilova mereni ve smeru Y}

procedure ZhladRychlosti(var Rychlosti: T3DDoubleArray; const vrstva: integer;
  const NrOfPass: integer); {zhladi rychlosti ve vrstve vrstva. Algoritmus je stejny
    jako pro interpolaci, probehne NrOfPass-krat}
procedure VytvorIdentifikatory(const Coord: TCoord; var Ident: T2DDoubleArray;
  var NrGph: integer);
  {Vytvori indexy geofonu - stejne geofony v zaznamu budou mit stejny index.
  Coord jsou souradnice geofonum, v poli Ident se vrati identifikatory geofonu,
  v NrGph vrati pocet geofonu}
procedure VyplnVectB(var VectB: TDoubleArray; const TimeData: TTimeData;
  const MatA: T2DDoubleArray);
  {Procedura naplni vektor pravych stran casy prichodu z TimeData}
procedure VyplnMatici(var MatA: T2DDoubleArray; const ShotCoord, GeophCoord:
  TCoord; const TimeData: TTimeData; const Rychlosti: T3DDoubleArray;
  const PocetVrstev, PocetGeofonu, PocetBunek: integer; const OdpGf: TOdpGf);
  {Procedura vyplni matici A}
procedure VyplnProfilovouMatici(var MatA: T2DDoubleArray; const ShotCoord, GeophCoord:
  TCoord; const TimeData: TTimeData; const Rychlosti: T3DDoubleArray;
  const PocetVrstev, PocetGeofonu: integer; const ProfilCoord: TDoubleArray;
  const OdpGf: TOdpGf);  {Vyplni Matici A pro profilove mereni}
procedure SpoctiCtverecky(const xs, ys, xg, yg: double;
  const Xsize, Ysize: integer; var vysledek: TCtverCoord); {Procedura zjisti
  pres ktere bunky prochazi paprsek z [xs, ys] do [xg, yg]. Vrati to v poli
  vysledek -- vysledek[i,0] a [i,1] jsou x a y souradnice bunky,
  vysledek[i,2] je vzdalenost, kterou to tou bunkou urazi.}
procedure UrciProfilCoord(const ShotCoord, GeophCoord: TCoord;
  var ProfilCoordTransp: TDoubleArray); {Spoji ShotCoord a GeophCoord, vyhazi
    duplicity. Vysledkem jsou serazene body mereni na profilu}
function delka(const x1, y1, x2, y2 : double):double;
Procedure SpoctiSoustavu(const MatA: T2DDoubleArray; const VectB: TDoubleArray;
  const M, N, ZacCtver, ZacOdp: integer; var VectX, VectStd: TDoubleArray; var RMS: double);
  {Solves the system of equations defined by MatA and VectB using the SVD routine}
procedure Zapis();
procedure ZapisMatA(const NrOdp: integer; const ZacatekOdp: integer;
  const PocetGeofonu: integer);
procedure ZapisVectB();
procedure ZapisVysledek(const fw_name: string);
procedure PrectiVectX(const XSize: integer);
procedure PrectiVectStd(const XSize: integer);
{procedures Swap and Sort Cut & Paste from Delphi Fundamental (cUtils) by
David Butler http://fundementals.sourceforge.net/}
Procedure Swap (var X, Y : double);
Procedure Sort (var Key : TDoubleArray; var Data : TDoubleArray); overload;
Procedure Sort (var V : TDoubleArray); overload;
{$IFDEF UNIX}
procedure RunCommand(Command: string);
procedure SpustSVD();
{$ENDIF}
procedure NajdiStejne(var OdpGf: TOdpGf; const ShotCoord, GeophCoord: TCoord;
  Identif: T2DDoubleArray; const TimeData: TTimeData); {Najde odpaly ktere jsou na
  stejnem miste jako geofony. V OdpGf[0] vrati cislo techto odpalu, v OdpGf[1]
  vrati identifikator  prislusneho geofonu}


function IsNumeric(c: char): Boolean; {copy&paste from IdGlobal}
function GetNrOfKnownVel(): Integer; {Funkce zjisti pocet znamych rychlosti
  zadanych v sgKnownVel}
function GetNrOfKnownDepths(): Integer; {Funkce zjisti pocet znamych hloubek
  zadanych v sgKnownDepths}

function RefrMaxValue(const Data: array of Double): Double;
function RefrMinValue(const Data: array of Double): Double;
function ReadRMS(const RmsFileName: string): Double;  {Funkce precte ze souboru
  hodnotu RMS}
procedure PseudoInverse(const Matr: TReal2DArray; const tol:double;
   var MatrInv: TReal2DArray);  {computes pseudoinverse of matrix Matr, tol is
   the lowest singular number allowed}

implementation

uses
  main;

procedure SpoctiVp(const TimeData: TTimeData; var Rychlosti: T3dDoubleArray;
  var VpZname: T2dIntegerArray; const ShotCoord, GeophCoord: TCoord);
var
  i: integer;
  sourad: integer;
  vp: double;
  vzd: double;
  x, y: integer;
begin
  for i := 0 to (PocetZaznamu - 1) do
    begin
      if i >= PocetZaznamu then
        MessageDlg('chyba', mtError, [mbAbort], 0);
      if TimeData[i, 2] = 1 then {pokud je to hodochrona prime vlny}
        begin
          sourad := trunc(TimeData[i, 0]);
          vzd := delka(ShotCoord[sourad, 0], ShotCoord[sourad, 1],
            GeophCoord[i, 0], GeophCoord[i, 1]);
          if abs(TimeData[i, 1]) < MinTime then continue;
          vp := vzd / TimeData[i, 1];  {rychlost prime vlny}
          x := round(ShotCoord[sourad, 0] + (GeophCoord[i, 0] -
            ShotCoord[sourad, 0]) / 2 - MinXT);
          y := round(ShotCoord[sourad, 1] + (GeophCoord[i, 1] -
            ShotCoord[sourad, 1]) / 2 - MinYT);
          Rychlosti[0, x, y] := vp;
          VpZname[x, y] := 1;
        end;
    end;
end;

procedure InterpolujVpXProfil(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double);
var
  i: integer;
  it : integer;
  vel: double;
  change, TmpChange: double;
begin
  it := 0;
  repeat
    TmpChange := 0;
    Change := 0;
    {rychlosti uprostred}
    for i := 1 to (NrX - 2) do
      if VpZname[i, 0] = 0 then
        begin
          vel := (Rychlosti[vrstva, i - 1, 0] + Rychlosti[vrstva, i + 1, 0])/2;
          TmpChange := abs(vel - Rychlosti[vrstva, i, 0]);
          Change := Max(Change, TmpChange);
          Rychlosti[vrstva, i, 0] := vel;
        end;
    {rychlost na levem okraji}
    if VpZname[0, 0] = 0 then
      begin
        vel := Rychlosti[vrstva, 1, 0];
        TmpChange := abs(vel - Rychlosti[vrstva, 0, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, 0] := vel;
      end;
    {rychlost na pravem okraji}
    if VpZname[NrX - 1, 0] = 0 then
      begin
        vel := Rychlosti[vrstva, NrX - 2, 0];
        TmpChange := abs(vel - Rychlosti[vrstva, NrX - 1, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, NrX - 1, 0] := vel;
      end;
    Inc(it);
  until (Change < MaxDiff) AND (it > 1);
  frmMain.StatusBar1.SimpleText := 'Interpolating v0, iteration ' + IntToStr(it);
end;

procedure InterpolujVpYProfil(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double);
var
  i: integer;
  it : integer;
  vel: double;
  change, TmpChange: double;
begin
  it := 0;
  repeat
    TmpChange := 0;
    Change := 0;
    {rychlosti uprostred}
    for i := 1 to (NrY - 2) do
      if VpZname[0, i] = 0 then
        begin
          vel := (Rychlosti[vrstva, 0, i - 1] + Rychlosti[vrstva, 0, i + 1])/2;
          TmpChange := abs(vel - Rychlosti[vrstva, 0, i]);
          Change := Max(Change, TmpChange);
          Rychlosti[vrstva, 0, i] := vel;
        end;
    {rychlost na levem okraji}
    if VpZname[0, 0] = 0 then
      begin
        vel := Rychlosti[vrstva, 0, 1];
        TmpChange := abs(vel - Rychlosti[vrstva, 0, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, 0] := vel;
      end;
    {rychlost na pravem okraji}
    if VpZname[NrY - 1, 0] = 0 then
      begin
        vel := Rychlosti[vrstva, 0, NrY - 2];
        TmpChange := abs(vel - Rychlosti[vrstva, 0, NrY - 1]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, NrY - 1] := vel;
      end;
    Inc(it);
  until (Change < MaxDiff) AND (it > 1);
  frmMain.StatusBar1.SimpleText := 'Interpolating v0, iteration ' + IntToStr(it);
end;

procedure InterpolujVp(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double);
var
  i, j: integer;
  it: integer;
  vel: double;
  change, TmpChange: double;
begin
  it := 0;
  repeat
  {rychlosti uprostred}
    TmpChange := 0;
    Change := 0;
    for i := 1 to (NrX - 2) do
      for j := 1 to (NrY - 2) do
        if VpZname[i, j] = 0 then
          begin
            vel := (Rychlosti[vrstva, i - 1, j] + Rychlosti[vrstva, i + 1, j] +
              Rychlosti[vrstva, i, j - 1] + Rychlosti[vrstva, i, j + 1]) / 4;
            TmpChange := abs(vel - Rychlosti[vrstva, i, j]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, i, j] := vel;
          end;
   {rychlosti na okrajich}
    for j := 1 to (NrY - 2) do
      begin
        if VpZname[0, j] = 0 then
          begin
            vel := (Rychlosti[vrstva, 1, j] +
              Rychlosti[vrstva, 0, j - 1] + Rychlosti[vrstva, 0, j + 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, 0, j]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, 0, j] := vel;
          end;
        if VpZname[NrX - 1, j] = 0 then
          begin
            vel := (Rychlosti[vrstva, NrX - 2, j] +
              Rychlosti[vrstva, NrX - 1, j - 1] +
              Rychlosti[vrstva, NrX - 1, j + 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, Nrx - 1, j]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, NrX - 1, j] := vel;
          end;
      end;
    for i := 1 to (NrX - 2) do
      begin
        if VpZname[i, 0] = 0 then
          begin
            vel := (Rychlosti[vrstva, 1, NrY - 2] +
              Rychlosti[vrstva, 0, NrY - 1] + Rychlosti[vrstva, 0, NrY - 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, i, 0]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, i, 0] := vel;
          end;
        if VpZname[i, NrY - 1] = 0 then
          begin
            vel := (Rychlosti[vrstva, i, NrY - 2] +
              Rychlosti[vrstva, i - 1, NrY - 1] +
              Rychlosti[vrstva, i + 1, NrY - 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, i, NrY - 1]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, i, NrY - 1] := vel;
          end;
      end;
    {rychlosti v rozich}
    if VpZname[0, 0] = 0 then
      begin
        vel := (Rychlosti[vrstva, 0, 1] + Rychlosti[vrstva, 1, 0]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, 0, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, 0] := vel;
      end;
    if VpZname[0, NrY - 1] = 0 then
      begin
        vel := (Rychlosti[vrstva, 0, NrY - 2] + Rychlosti[vrstva, 1, NrY - 1]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, 0, NrY - 1]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, NrY - 1] := vel;
      end;
    if VpZname[NrX - 1, 0] = 0 then
      begin
        vel := (Rychlosti[vrstva, NrX - 2, 0] + Rychlosti[vrstva, NrX - 1, 1]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, NrX - 1, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, NrX - 1, 0] := vel;
      end;
    if VpZname[NrX - 1, NrY - 1] = 0 then
      begin
        vel := (Rychlosti[vrstva, NrX - 1, NrY - 2] +
          Rychlosti[vrstva, NrX - 2, NrY - 1]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, NrX - 1, NrY - 1]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, NrX - 1, nrY - 1] := vel;
      end;
    Inc(it);
  until (Change < MaxDiff) AND (it > 1);
  frmMain.StatusBar1.SimpleText := 'Interpolating v0, iteration ' + IntToStr(it);
end;

procedure InterpolujV0(var Rychlosti: T3dDoubleArray; const VpZname: T2DIntegerArray;
  const vrstva: integer; const MaxDiff: double);
var
  i, j: integer;
  it: integer;
  vel: double;
  change, TmpChange: double;
begin
  it := 0;
  repeat
  {rychlosti uprostred}
    TmpChange := 0;
    Change := 0;
    for i := 1 to (NrX - 2) do
      for j := 1 to (NrY - 2) do
        if Rychlosti[vrstva, i, j] = 1 then
          begin
            vel := (Rychlosti[vrstva, i - 1, j] + Rychlosti[vrstva, i + 1, j] +
              Rychlosti[vrstva, i, j - 1] + Rychlosti[vrstva, i, j + 1]) / 4;
            TmpChange := abs(vel - Rychlosti[vrstva, i, j]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, i, j] := vel;
          end;
   {rychlosti na okrajich}
    for j := 1 to (NrY - 2) do
      begin
        if Rychlosti[vrstva, 0, j] = 1 then
          begin
            vel := (Rychlosti[vrstva, 1, j] +
              Rychlosti[vrstva, 0, j - 1] + Rychlosti[vrstva, 0, j + 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, 0, j]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, 0, j] := vel;
          end;
        if Rychlosti[vrstva, NrX - 1, j] = 1 then
          begin
            vel := (Rychlosti[vrstva, NrX - 2, j] +
              Rychlosti[vrstva, NrX - 1, j - 1] +
              Rychlosti[vrstva, NrX - 1, j + 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, Nrx - 1, j]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, NrX - 1, j] := vel;
          end;
      end;
    for i := 1 to (NrX - 2) do
      begin
        if Rychlosti[vrstva, i, 0] = 1 then
          begin
            vel := (Rychlosti[vrstva, 1, NrY - 2] +
              Rychlosti[vrstva, 0, NrY - 1] + Rychlosti[vrstva, 0, NrY - 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, i, 0]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, i, 0] := vel;
          end;
        if Rychlosti[vrstva, i, NrY - 1] = 1 then
          begin
            vel := (Rychlosti[vrstva, i, NrY - 2] +
              Rychlosti[vrstva, i - 1, NrY - 1] +
              Rychlosti[vrstva, i + 1, NrY - 1]) / 3;
            TmpChange := abs(vel - Rychlosti[vrstva, i, NrY - 1]);
            Change := Max(Change, TmpChange);
            Rychlosti[vrstva, i, NrY - 1] := vel;
          end;
      end;
    {rychlosti v rozich}
    if Rychlosti[vrstva, 0, 0] = 1 then
      begin
        vel := (Rychlosti[vrstva, 0, 1] + Rychlosti[vrstva, 1, 0]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, 0, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, 0] := vel;
      end;
    if Rychlosti[vrstva, 0, NrY - 1] = 1 then
      begin
        vel := (Rychlosti[vrstva, 0, NrY - 2] + Rychlosti[vrstva, 1, NrY - 1]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, 0, NrY - 1]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, 0, NrY - 1] := vel;
      end;
    if Rychlosti[vrstva, NrX - 1, 0] = 1 then
      begin
        vel := (Rychlosti[vrstva, NrX - 2, 0] + Rychlosti[vrstva, NrX - 1, 1]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, NrX - 1, 0]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, NrX - 1, 0] := vel;
      end;
    if Rychlosti[vrstva, NrX - 1, NrY - 1] = 1 then
      begin
        vel := (Rychlosti[vrstva, NrX - 1, NrY - 2] +
          Rychlosti[vrstva, NrX - 2, NrY - 1]) / 2;
        TmpChange := abs(vel - Rychlosti[vrstva, NrX - 1, NrY - 1]);
        Change := Max(Change, TmpChange);
        Rychlosti[vrstva, NrX - 1, nrY - 1] := vel;
      end;
    Inc(it);
  until (Change < MaxDiff) AND (it > 1);
  frmMain.StatusBar1.SimpleText := 'Iterace ' + IntToStr(it);
end;

procedure ZhladRychlosti(var Rychlosti: T3DDoubleArray; const vrstva: integer;
  const NrOfPass: integer);
var
  i, j: integer;
  vel: double;
  it: integer;
begin
  if (NrX > 1) AND (NrY > 1) then
    begin
      for it := 0 to (NrOfPass - 1) do
        begin
        {rychlosti uprostred}
          for i := 1 to (NrX - 2) do
            for j := 1 to (NrY - 2) do
              begin
               vel := (Rychlosti[vrstva, i - 1, j] + Rychlosti[vrstva, i + 1, j] +
                 Rychlosti[vrstva, i, j - 1] + Rychlosti[vrstva, i, j + 1]) / 4;
               Rychlosti[vrstva, i, j] := vel;
             end;
         {rychlosti na okrajich}
          for j := 1 to (NrY - 2) do
            begin
              vel := (Rychlosti[vrstva, 1, j] +
                Rychlosti[vrstva, 0, j - 1] + Rychlosti[vrstva, 0, j + 1]) / 3;
              Rychlosti[vrstva, 0, j] := vel;

              vel := (Rychlosti[vrstva, NrX - 2, j] +
                Rychlosti[vrstva, NrX - 1, j - 1] +
                Rychlosti[vrstva, NrX - 1, j + 1]) / 3;
              Rychlosti[vrstva, NrX - 1, j] := vel;
            end;
          for i := 1 to (NrX - 2) do
            begin
              vel := (Rychlosti[vrstva, i, 1] +
                Rychlosti[vrstva, i - 1, 0] + Rychlosti[vrstva, i + 1, 0]) / 3;
              Rychlosti[vrstva, i, 0] := vel;

              vel := (Rychlosti[vrstva, i, NrY - 2] +
                Rychlosti[vrstva, i - 1, NrY - 1] +
                Rychlosti[vrstva, i + 1, NrY - 1]) / 3;
              Rychlosti[vrstva, i, NrY - 1] := vel;
            end;
          {rychlosti v rozich}
          vel := (Rychlosti[vrstva, 0, 1] + Rychlosti[vrstva, 1, 0]) / 2;
          Rychlosti[vrstva, 0, 0] := vel;

          vel := (Rychlosti[vrstva, 0, NrY - 2] + Rychlosti[vrstva, 1, NrY - 1]) / 2;
          Rychlosti[vrstva, 0, NrY - 1] := vel;

          vel := (Rychlosti[vrstva, NrX - 2, 0] + Rychlosti[vrstva, NrX - 1, 1]) / 2;
          Rychlosti[vrstva, NrX - 1, 0] := vel;

          vel := (Rychlosti[vrstva, NrX - 1, NrY - 2] +
            Rychlosti[vrstva, NrX - 2, NrY - 1]) / 2;
          Rychlosti[vrstva, NrX - 1, nrY - 1] := vel;
        end;
    end; {of if (NrX > 1) AND (NrY > 1) then}
  if (NrX = 1) AND (NrY > 1) then
    begin
      for it := 0 to (NrOfPass - 1) do
        begin
          {rychlosti uprostred}
          for i := 1 to (NrY - 2) do
            begin
              vel := (Rychlosti[vrstva, 0, i] + Rychlosti[vrstva, 0, i - 1] +
                Rychlosti[vrstva, 0 , i + 1]) / 3;
              Rychlosti[vrstva, 0, i] := vel;
            end;
          {rychlosti na okraji}
          vel := (Rychlosti[vrstva, 0, 0] + Rychlosti[vrstva, 0, 1]) / 2;
          Rychlosti[vrstva, 0 ,0] := vel;
          vel := (Rychlosti[vrstva, 0, NrY - 1] + Rychlosti[vrstva, 0, NrY - 2]) / 2;
          Rychlosti[vrstva, 0, NrY - 1] := vel;
        end; {of     for it := 0 to (NrOfPass - 1) do}
    end; {of if (NrX = 1) AND (NrY > 1) then}
  if (NrX > 1) AND (NrY = 1) then
    begin
      for it := 0 to (NrOfPass - 1) do
        begin
          {rychlosti uprostred}
          for i := 1 to (NrX - 2) do
            begin
              vel := (Rychlosti[vrstva, i, 0] + Rychlosti[vrstva, i - 1, 0] +
                Rychlosti[vrstva, i + 1, 0]) / 3;
              Rychlosti[vrstva, i, 0] := vel;
            end;
          {rychlosti na okraji}
          vel := (Rychlosti[vrstva, 0, 0] + Rychlosti[vrstva, 1, 0]) / 2;
          Rychlosti[vrstva, 0 ,0] := vel;
          vel := (Rychlosti[vrstva, NrX - 1, 0] + Rychlosti[vrstva, NrX - 2, 0]) / 2;
          Rychlosti[vrstva, NrX - 1, 0] := vel;
        end; {of     for it := 0 to (NrOfPass - 1) do}
    end; {of if (NrX > 1) AND (NrY = 1) then}

end;


function delka(const x1, y1, x2, y2 : double):double;
begin
  result := sqrt(sqr(x1 - x2) + sqr(y1 - y2));
end;


//procedure VytvorIdentifikatory(const Coord: TCoord; var Ident: TDoubleArray;
  //var NrGph: integer);
//var
  //i, j: integer;
  //id: integer;
  //x, y: double;
  //tmpIdent: array [0..500] of double;
  //tmpCoord: array [0..500] of array [0..1] of double;
//begin
  //for i := 0 to (PocetZaznamu - 1) do
    //Ident[i] := -1;
  //for i:= 0 to (pocetZaznamu - 1) do
    //begin
      //tmpCoord[i,0] := Coord[i, 0];
      //tmpCoord[i,1] := Coord[i, 1];
    //end;
  //id := 0;
  //for i := 0 to (PocetZaznamu - 1) do
    //begin
      //tmpIdent[i] := Ident[i];
      //if Ident[i] = -1 then
        //begin
          //x := Coord[i, 0];
          //y := Coord[i, 1];
          //for j := 0 to (PocetZaznamu - 1) do
            //begin
              //if (Coord[j, 0] = x) AND (Coord[j, 1] = y) then
                //begin
                  //Ident[j] := id;
                  //tmpIdent[j] := Ident[j];
                //end;
            //end;
          //Inc(id);
        //end;
    //end;
  //NrGph := id; {id zacina nulou, ale na konci se jeste jednou zvetsi, tak
    //neni treba nic pricitat/odcitat}

//end;


procedure VytvorIdentifikatory(const Coord: TCoord; var Ident: T2DDoubleArray;
  var NrGph: integer);
var
  i, j: integer;
  id: integer;
  x, y: double;
begin
  for i := 0 to (PocetZaznamu - 1) do
    Ident[i, 0] := -1;
  id := 0;
  for i := 0 to (PocetZaznamu - 1) do
    begin
      if Ident[i, 0] = -1 then
        begin
          x := Coord[i, 0];
          y := Coord[i, 1];
          for j := 0 to (PocetZaznamu - 1) do
            begin
              if (Coord[j, 0] = x) AND (Coord[j, 1] = y) then
                begin
                  Ident[j, 0] := i;
                  Ident[j, 1] := id;
                end;
            end;
          Inc(id);
        end;
    end;
  NrGph := id; {id zacina nulou, ale na konci se jeste jednou zvetsi, tak
    neni treba nic pricitat/odcitat}

end;

procedure VyplnVectB(var VectB: TDoubleArray; const TimeData: TTimeData;
  const MatA: T2DDoubleArray);
var
  i, j: integer;
  posun: integer;
  nula: Boolean;
begin
  for i := 0 to (Length(TimeData) - 1) do
    begin
      {nutno stanovit, kdy jsou vsechny prvky radku MatA nulove, potom musi byt
       nula i prislusny prvek VectB}
      nula := TRUE;
      for j := 0 to (Length(MatA) - 1) do
        begin
          if MatA[j, i] <> 0 then nula := False;
        end;
      if nula = False then
        VectB[i] := TimeData[i, 1]
      else
        VectB[i] := 0;
    end;

  posun := Length(TimeData);
  for i := 0 to (Length(OdpGf) - 1) do
    VectB[i + posun] := 0;
  posun := Length(TimeData) + Length(OdpGf);
  for i := 0 to (NrOfKnownVel - 1) do
    VectB[i + posun] := 1 / StrToFloat(frmMain.sgKnownVel.Cells[3, i + 1]);
end;

procedure VyplnProfilovouMatici(var MatA: T2DDoubleArray; const ShotCoord, GeophCoord:
  TCoord; const TimeData: TTimeData; const Rychlosti: T3DDoubleArray;
  const PocetVrstev, PocetGeofonu: integer; const ProfilCoord: TDoubleArray;
  const OdpGf: TOdpGf);
var
  i, j, k: integer;
  odpal, geofon: integer;
  ZacatekGf, ZacatekCtver: integer;
  vrstva: integer;
  vzd: double;
  PocUseku: integer;
  SourOdp, SourGf: integer;
  mensi, vetsi: integer;
  NrOdp: integer;
  xs, ys, xg, yg: integer;
  zdroj, prijimac: double;
  x: double;
  ZacatekRychlosti: integer;
begin
  NrOdp := Length(ShotCoord);
  ZacatekGf :=  NrOdp * (PocetVrstev - 1);
  ZacatekCtver := NrOdp * (PocetVrstev - 1) + PocetGeofonu * (PocetVrstev - 1);
  PocUseku := Length(ProfilCoord) - 1;
  for i := 0 To (Length(TimeData) - 1) do
    begin
      odpal := trunc(TimeData[i, 0]);
      geofon := trunc(TimeData[i, 3]);
      vrstva := trunc(TimeData[i, 2]);
      for j :=  0 to (vrstva - 2) do
        begin
          xs := round(ShotCoord[odpal, 0] - MinXT);
//          zdroj := (ShotCoord[odpal, 0] - MinXT);
          zdroj := ShotCoord[odpal, 0];
//          ys := round(ShotCoord[odpal, 1] - MinYT);
          ys := 0;
          if (Rychlosti[j, xs, ys] <> 0) AND (abs(Rychlosti[j, xs, ys]) <
            abs(Rychlosti[j + 1, xs, ys])) then
            MatA[odpal + j * NrOdp, i] := 1 / Rychlosti[j, xs, ys] *
              sqrt(1 - sqr(Rychlosti[j, xs, ys] / Rychlosti[j + 1, xs, ys]))
          else
            MatA[odpal + j * NrOdp, i] := 0;
          xg := round(GeophCoord[geofon, 0] - MinXT);
//          prijimac := (GeophCoord[geofon, 0] - MinXT);
          prijimac := GeophCoord[geofon, 0];
//          yg := round(GeophCoord[geofon, 1] - MinYT);
          yg := 0;
          if (Rychlosti[j, xg, yg] <> 0) AND (abs(Rychlosti[j, xg, yg]) <
            abs(Rychlosti[j + 1, xg, yg])) then
            MatA[ZacatekGf + geofon + j * PocetGeofonu, i] := 1 /
              Rychlosti[j, xg, yg] * sqrt(1 - sqr(Rychlosti[j, xg, yg] /
              Rychlosti[j + 1, xg, yg]))
          else
            MatA[ZacatekGf + geofon + j * PocetGeofonu, i] := 0;


          k := 0;
          while zdroj <> ProfilCoord[k] do Inc(k);
          SourOdp := k;
          k := 0;
          while prijimac <> ProfilCoord[k] do Inc(k);
          SourGf := k;
          mensi := Min(SourOdp, SourGf);
          vetsi := Max(SourOdp, SourGf);

          for k := mensi to (vetsi - 1) do
            begin
              vzd := ProfilCoord[k + 1] - ProfilCoord[k];
              MatA[ZacatekCtver + PocUseku * j + k, i] := vzd;
            end;
        end {of for-j};
    end {of for-i};
  for i := 0 to (Length(OdpGf) - 1) do
    for j := 0 to (OdpGf[i, 2] - 2) do
      begin
        MatA[OdpGf[i, 0] + NrOdp * j, i + Length(TimeData)] := 1;
        MatA[ZacatekGf + OdpGf[i, 1] + PocetGeofonu * j, i + Length(TimeData)] := -1;
      end;
  ZacatekRychlosti := Length(TimeData) + Length(OdpGf);
  for i := 1 to NrOfKnownVel do
    begin
      j := trunc(StrToFloat(frmMain.sgKnownVel.Cells[i, 4])) - 1;
      x := StrToFLoat(frmMain.sgKnownVel.Cells[i, 1]) / Scale - MinXT;
//      y := frmMain.sgKnownVel.Cells[i, 2] / Scale - MinYT;
      MatA[ZacatekCtver + j * PocetGeofonu + trunc(x), i + ZacatekRychlosti] := 1;
    end;
end;

procedure VyplnMatici(var MatA: T2DDoubleArray; const ShotCoord, GeophCoord:
  TCoord; const TimeData: TTimeData; const Rychlosti: T3DDoubleArray;
  const PocetVrstev, PocetGeofonu, PocetBunek: integer; const OdpGf: TOdpGf);
var
  i, j, k: integer;
  odpal, geofon, GfId: integer;
  NrOdp: integer;
  xs, ys, xg, yg: integer;
  ZacatekGf, ZacatekCtver: integer;
  vrstva: integer;
//  PocCtver: integer;
  ctverecky: TCtverCoord; {v prvnim sloupci je cislo ctverecku (bunky), v druhem
    vzdalenost, jakou v nem paprsek urazi}
  x, y: double;
  ZacatekRychlosti: integer;
begin
  NrOdp := Length(ShotCoord);
  ZacatekGf :=  NrOdp * (PocetVrstev - 1);
  ZacatekCtver := NrOdp * (PocetVrstev - 1) + PocetGeofonu * (PocetVrstev - 1);

  for i := 0 To (Length(TimeData) - 1) do
    begin
      odpal := trunc(TimeData[i, 0]);
      geofon := trunc(TimeData[i, 3]);
      GfId := trunc(TimeData[i, 4]);
      vrstva := trunc(TimeData[i, 2]);
      for j :=  0 to (vrstva - 2) do
        begin
//          xs := round(ShotCoord[odpal, 0] - MinXT);
//          ys := round(ShotCoord[odpal, 1] - MinYT);
          xs := trunc(ShotCoord[odpal, 0] - MinXT);
          ys := trunc(ShotCoord[odpal, 1] - MinYT);
          if (Rychlosti[j, xs, ys] <> 0) AND (abs(Rychlosti[j, xs, ys]) <
            abs(Rychlosti[j + 1, xs, ys])) then
            MatA[odpal + j * NrOdp, i] := (1 / Rychlosti[j, xs, ys]) *
              sqrt(1 - sqr(Rychlosti[j, xs, ys] / Rychlosti[j + 1, xs, ys]))
          else
            MatA[odpal + j * NrOdp, i] := 0;
//          xg := round(GeophCoord[geofon, 0] - MinXT);
//          yg := round(GeophCoord[geofon, 1] - MinYT);
          xg := trunc(GeophCoord[geofon, 0] - MinXT);
          yg := trunc(GeophCoord[geofon, 1] - MinYT);
          if (Rychlosti[j, xg, yg] <> 0) AND (abs(Rychlosti[j, xg, yg]) <
            abs(Rychlosti[j + 1, xg, yg])) then
            MatA[ZacatekGf + GfId + j * PocetGeofonu, i] := (1 /
              Rychlosti[j, xg, yg]) * sqrt(1 - sqr(Rychlosti[j, xg, yg] /
              Rychlosti[j + 1, xg, yg]))
          else
            MatA[ZacatekGf + GfId + j * PocetGeofonu, i] := 0;
{          SpoctiCtverecky(xs, ys, xg, yg, NrX, NrY, Ctverecky);}
          SpoctiCtverecky(ShotCoord[odpal, 0] - MinXT, ShotCoord[odpal, 1] - MinYT,
            GeophCoord[geofon, 0] - MinXT, GeophCoord[geofon, 1] - MinYT, NrX,
            NrY, Ctverecky);
{          for k := 0 to (Length(Ctverecky) - 1) do
            if (Rychlosti[j, trunc(Ctverecky[k, 0]), trunc(Ctverecky[k, 1])] <> 0) then
              MatA[ZacatekCtver + trunc(Ctverecky[k, 3]) + j * PocetBunek, i] :=
                1 / Rychlosti[j, trunc(Ctverecky[k, 0]), trunc(Ctverecky[k, 1])] *
                Ctverecky[k, 2]
            else
              MatA[ZacatekCtver + trunc(Ctverecky[k, 3]) + j * PocetBunek, i] := 0;}
          for k := 0 to (Length(Ctverecky) - 1) do
            if Ctverecky[k, 2] <> 0 then {pokud je zdroj nebo geofon na rozhrani,
              vyjde delka pruchodu posledni bunkou 0 a prepsalo by to predtim
              spoctenou vzdalenost}
              begin
                if Ctverecky[k, 2] > NumZero then
                  MatA[ZacatekCtver + trunc(Ctverecky[k, 3]) + j * PocetBunek, i] :=
                    Ctverecky[k, 2]
                else
                  MatA[ZacatekCtver + trunc(Ctverecky[k, 3]) + j * PocetBunek, i] :=
                    0;
              end {of for-k};

        end {of for-j};
    end {of for-i};
  for i := 0 to (Length(OdpGf) - 1) do
//    for j := 0 to (OdpGf[i, 2] - 2) do
    for j := 0 to (PocetVrstev - 2) do
      begin
        MatA[OdpGf[i, 0] + NrOdp * j, i + Length(TimeData)] := 1;
        MatA[ZacatekGf + OdpGf[i, 1] + PocetGeofonu * j, i + Length(TimeData)] := -1;
      end;
  ZacatekRychlosti := Length(TimeData) + Length(OdpGf);
  for i := 1 to NrOfKnownVel do
    begin
      j := trunc(StrToFloat(frmMain.sgKnownVel.Cells[4, i])) - 2;
      x := StrToFloat(frmMain.sgKnownVel.Cells[1, i]) / Scale - MinXT;
      y := StrToFloat(frmMain.sgKnownVel.Cells[2, i]) / Scale - MinYT;
//      MatA[ZacatekCtver + j * PocetBunek + trunc(y - 1) * NrX + trunc(x),
//        i - 1 + ZacatekRychlosti] := 1;
      MatA[ZacatekCtver + j * PocetBunek + trunc(y) * NrX + trunc(x),
        i - 1 + ZacatekRychlosti] := 1;
    end;

{  for i := 1 to NrOfKnownDepth do
    begin
      j := trunc(StrToFloat(frmMain.sgKnownDepth.Cells[4, i])) - 2;
      x := StrToFloat(frmMain.sgKnownDepth.Cells[1, i]) / Scale - MinXT;
      y := StrToFloat(frmMain.sgKnownDepth.Cells[2, i]) / Scale - MinYT;
//      MatA[ZacatekCtver + j * PocetBunek + trunc(y - 1) * NrX + trunc(x),
//        i - 1 + ZacatekRychlosti] := 1;
      MatA[ZacatekCtver + j * PocetBunek + trunc(y) * NrX + trunc(x),
        i - 1 + ZacatekRychlosti] := 1;
    end;
}
end;


procedure SpoctiCtverecky(const xs, ys, xg, yg: double;
  const Xsize, Ysize: integer; var vysledek: TCtverCoord);
var
//  pomdelka: double;
  k, q: double;
  i: integer;
  PocBun: integer; {pocet protnutych bunek}
  PocSour: integer; {pocet protnutych hranic bunek}
  souradnice: TCoord;
  XStep, YStep: integer;
  CurX, CurY: double;
  PocetSpoctenych: integer;
  NextIt: boolean;
  kInf: boolean;
//  str: string;
  pomx, pomy: TDoubleArray;
  index: integer;
begin
  kInf := False;
//  pomdelka := delka(xs, ys, xg, yg);
//  m := 0;
//  if ((xs - xg) > 1) then
//    begin
      if abs(xs - xg) > NumZero then
        k := (ys - yg) / (xs - xg)
      else
        kInf := True; {koeficient k je nekonecno -- primka jde ve smeru osy y}
      q := yg - k * xg;
{      if (abs((Trunc(xs) - trunc(xg))) <= NumZero) then
        begin
          str := 'Source X coordinate [' + FloatToStr(xs) + ';' +
            FloatToStr(ys) + '] is too close to reciever X coordinate [' +
            FloatToStr(xg) + ';' + FloatToStr(yg) + ']. Closer than ' +
            FloatToStr(NumZero);
          raise Exception.Create(str);
        end;
      if (abs((Trunc(ys) - trunc(yg))) <= NumZero) then
        begin
          str := 'Source Y coordinate [' + FloatToStr(xs) + ';' +
            FloatToStr(ys) + '] is too close to reciever Y coordinate [' +
            FloatToStr(xg) + ';' + FloatToStr(yg) + ']. Closer than ' +
            FloatToStr(NumZero);
          raise Exception.Create(str);
        end;
}
      PocBun := abs(trunc(xs) - trunc(xg)) + abs(trunc(ys) - trunc(yg)) + 1;
        {Pocet bunek, kterymi paprsek projde je o jednu vetsi, nez pocet
        protnutych hranic}
      SetLength(Vysledek, PocBun);
      PocSour := abs(trunc(xs) - trunc(xg)) + abs(trunc(ys) - trunc(yg));
      {i souradnice [xs,ys], [xg,yg] <-- po uprave uz neplati}
      SetLength(souradnice, PocSour);
      if xs > xg then
        XStep := -1
      else
        XStep := 1;
      if ys > yg then
        YStep := -1
      else
        YStep := 1;

      PocetSpoctenych := -1;
      {Ted pocitame, kde protne paprsek X-hranici bunek (kde je X cele cislo)}
      if XStep = 1 then
        CurX := trunc(xs) + 1
      else
        CurX := trunc(xs);
      if (KInf = False) AND ((trunc(xs) - trunc(xg)) <> 0) then
        NextIt := True
      else
        NextIt := False; {pokud maji xs a xg stejnou x souradnici, tak zadnou
        x-linii neprotinaj}
      while NextIt = True do
        begin
          Inc(PocetSpoctenych);
          souradnice[PocetSpoctenych, 0] := CurX;
          souradnice[PocetSpoctenych, 1] := k * CurX + q;
          CurX := CurX + Xstep;
          if XStep = 1 then
            begin
              if CurX <= xg then
                NextIt := True
              else
                NextIt := False;
            end
          else
            begin
              if CurX > xg then
                NextIt := True
              else
                NextIt := False;
            end;
        end;
      {Ted pocitame, kde protne paprsek Y-hranici bunek (kde je Y cele cislo)}
      if YStep = 1 then
        CurY := trunc(ys) + 1
      else
        CurY := trunc(ys);
      if (trunc(ys) - trunc(yg)) <> 0 then
        NextIt := TRUE
      else
        NextIt := False; {pokud maji ys a yg stejnou y souradnici, tak zadnou
        y-linii neprotinaj}
      while NextIt = True do
        begin
          Inc(PocetSpoctenych);
          souradnice[PocetSpoctenych, 1] := CurY;
          if (abs(k) > NumZero) AND (kInf = False) then
            souradnice[PocetSpoctenych, 0] := (CurY - q) / k
          else
            souradnice[PocetSpoctenych, 0] := xs;
          CurY := CurY + Ystep;
          if YStep = 1 then
            begin
              if CurY <= yg then
                NextIt := True
              else
                NextIt := False;
            end
          else
            begin
              if CurY > yg then
                NextIt := True
              else
                NextIt := False;
            end;
        end;

        Inc(PocetSpoctenych); {pocet spoctenych je o jednu vetsi, zacinali jsme
          pocitat od nuly}
        if PocetSpoctenych <> (PocBun - 1) then
          MessageDlg('Chyba: PocetSpoctenych <> PocBun.', mtError, [mbAbort], 0);
        SetLength(pomx, PocetSpoctenych);
        SetLength(pomy, PocetSpoctenych);
        for i := 0 to (PocetSpoctenych - 1) do
          begin
            pomx[i] := souradnice[i, 0];
            pomy[i] := souradnice[i, 1];
          end;
        {seradime vypoctene souradnice pruseciku}
        if KInf = False then
          begin
            Sort(pomx, pomy); {pokud nejde primka ve smeru osy y, setridime
              podle x}
            if XStep = -1 then {jdeme do nizsich metrazi}
              for i := 0 to (PocetSpoctenych - 1) do
                begin
                  souradnice[i, 0] := pomx[PocetSpoctenych - 1 -i];
                  souradnice[i, 1] := pomy[PocetSpoctenych - 1 -i];
                end
            else    {jdeme do vyssich metrazi}
              for i := 0 to (PocetSpoctenych - 1) do
                begin
                  souradnice[i, 0] := pomx[i];
                  souradnice[i, 1] := pomy[i];
                end
          end
        else
          begin
            Sort(pomy, pomx); {jinak tridime podle y}
            if YStep = -1 then {jdeme do nizsich metrazi}
              for i := 0 to (PocetSpoctenych - 1) do
                begin
                  souradnice[i, 0] := pomx[PocetSpoctenych - 1 -i];
                  souradnice[i, 1] := pomy[PocetSpoctenych - 1 -i];
                end
            else    {jdeme do vyssich metrazi}
              for i := 0 to (PocetSpoctenych - 1) do
                begin
                  souradnice[i, 0] := pomx[i];
                  souradnice[i, 1] := pomy[i];
                end
          end;



//        vysledek[0, 0] := trunc(ys) * XSize + trunc(xs);
//        vysledek[0, 1] := delka(xs, ys, souradnice[0, 0], souradnice[0, 1]);
        vysledek[0, 0] := xs;
        vysledek[0, 1] := ys;
        if PocSour = 0 then
          vysledek[0, 2] := delka(xs, ys, xg, yg)
        else
          vysledek[0, 2] := delka(xs, ys, souradnice[0, 0], souradnice[0, 1]);
        vysledek[0, 3] := trunc(ys) * XSize + trunc(xs);
        index := 0;
        for i := 1 to (PocetSpoctenych -1) do
          begin
{            vysledek[i, 0] := trunc(souradnice[i, 1]) * XSize +
              trunc(souradnice[i, 0]);
            vysledek[i, 1] := delka(souradnice[i - 1, 0], souradnice[i - 1, 1],
              souradnice[i, 0], souradnice[i, 1]);
}
            vysledek[i, 0] := souradnice[i, 0];
            vysledek[i, 1] := souradnice[i, 1];
            vysledek[i, 2] := delka(souradnice[i - 1, 0], souradnice[i - 1, 1],
              souradnice[i, 0], souradnice[i, 1]);
            vysledek[i, 3] := trunc(souradnice[i - 1, 1]) * XSize +
              trunc(souradnice[i, 0]);
            index := i;
          end;
{        vysledek[PocBun - 1, 0] := trunc(yg) * XSize + trunc(xg);
        vysledek[PocBun - 1, 1] := delka(souradnice[PocSour - 1, 0],
          souradnice[PocSour - 1, 1], xg, yg);
}
        if PocSour > 0 then
          begin
            vysledek[PocBun - 1, 0] := xg;
            vysledek[PocBun - 1, 1] := yg;
            vysledek[PocBun - 1, 2] := delka(souradnice[PocSour - 1, 0],
              souradnice[PocSour - 1, 1], xg, yg);
    //        vysledek[PocBun - 1, 3] := trunc(yg) * XSize + trunc(xg);
            vysledek[PocBun - 1, 3] := trunc(souradnice[index, 1]) * XSize + trunc(xg);
         end;
end;

Procedure SpoctiSoustavu(const MatA: T2DDoubleArray; const VectB: TDoubleArray;
  const M, N, ZacCtver, ZacOdp: integer; var VectX, VectStd: TDoubleArray; var RMS:double);
  {Solves the system of equations defined by MatA and VectB using the SVD routine}
var
//pro SVD
//     W: TReal1DArray;
//     U: TReal2DArray;
//     VT: TReal2DArray;
     svdA: TReal2DArray; {matrix A of the set of equations for the SVD routine}
     svdATr: TReal2DArray; {transpose of matrix A}
//     svdB: TReal1DArray; {vector of right hand sides for the SVD routine}
//     M: integer; {x-size of the matrix A}
//     N: integer; {y-size of the matrix A and size of the VectB}
     mpr: TReal1DArray; {vector of initial depth and velocities}
     Cm: TReal2DArray; {covariance matrix of uncertainities of the input model}
     Cd: TReal2DArray; {covariance matrix of uncertainities of the measured data}
     Cmi: TReal2DArray; {inverse matrix to Cm}
     Cdi: TReal2DArray; {inverse matrix to Cd}
     i, j: integer;
     tmp, tmp1: TReal2DArray;
     work, tmpVect: TReal1DArray;
     InnerPart: TReal2DArray;
     Brack: TReal1DArray;
     suma: double;
  
begin
{in this part, we're going to solve the system of linear equations, using
the SVD from the ALGLIB collection -- http://www.alglib.net}

{the system of equations is solved using the following (Tarantola, Inverse Problem
Theory):

x = mpr + (A' * Cdi * A + Cmi)^-1 * A' * Cdi * (b - A * mpr)

}

//  frmMain.StatusBar1.SimpleText := 'Start of SVD.';
//  Application.ProcessMessages;

  try
    SetLength(svdA, N + 1, M + 1); {The matrices in ALGLIB begin indexing with 1,
      not 0 as standard dynamic arrays do.}

    for i := 0 to (M - 1) do
      for j := 0 to  (N - 1) do
        svdA[j + 1, i + 1] := MatA[i, j];

{    SetLength(svdB, N + 1);
    for i := 0 to (N - 1) do
      svdB[i + 1] := VectB[i];}

    SetLength(mpr, M + 1);
    mpr[0] := 0;
    SetLength(Cm, M + 1, M + 1);
    SetLength(Cmi, M + 1, M + 1);
    for i := 0 to M do
      for j := 0 to M do
        Cm[i, j] := 0;
        
    for i := 1 to ZacCtver do
      mpr[i] := depth_prior;
    for i := (ZacCtver + 1) to M do
      mpr[i] := 1 / vel_prior;

    for i := 1 to ZacCtver do
      Cm[i,i] := Power(depth_uncert, 2);
    for i := (ZacCtver + 1) to M do
      Cm[i, i] := Power(vel_uncert, 2);

    if Iterace > 0 then  {if we have already computed previous iteration, use
      selected results as an initial estimate}
      begin
        for i := (ZacCtver + 1) to M do
          if (VectStd[i - 1] / VectX[i - 1]) < err then
            begin
              mpr[i] := VectX[i - 1];
//              Cm[i, i] := VectStd[i - 1];
            end
      end;


    SetLength(Cd, N + 1, N + 1);
    SetLength(Cdi, N + 1, N + 1);
    for i := 0 to N do
      for j := 0 to N do
        Cd[i, j] := 0;
    for i := 1 to N do
      Cd[i,i] := Power(data_uncert, 2);

    {pseudoinverse of Cm}
    PseudoInverse(Cm, tol, Cmi);
    {pseudoinverse of Cd}
    PseudoInverse(Cd, tol, Cdi);

    {computing inner part of the solution -- (A' * Cdi * A + Cmi)^-1 }
    SetLength(svdATr, M + 1, N + 1);
    CopyAndTranspose(svdA, 1, N, 1, M, svdATr, 1, M, 1, N);
    {setting of temporary arrays}
    SetLength(tmp, M + 1, N + 1);
    SetLength(work, Max(M, N) + 1);
    MatrixMatrixMultiply(svdATr, 1, M, 1, N, False, Cdi, 1, N, 1, N, False, 1, tmp,
      1, M, 1, N, 0, work);
    SetLength(tmp1, M + 1, M + 1);
    MatrixMatrixMultiply(tmp, 1, M, 1, N, False, svdA, 1, N, 1, M, False, 1, tmp1,
      1, M, 1, M, 0, work);
    SetLength(work, 0);
    SetLength(tmp, M + 1, M + 1);
    for i := 1 to M do
      for j := 1 to M do
        tmp[i, j] := tmp1[i, j] + Cmi[i, j];
    SetLength(tmp1, 0, 0);
    SetLength(InnerPart, M + 1, M + 1);
    PseudoInverse(tmp, tol, InnerPart);
    SetLength(tmp1, 0, 0);

    {computing the brackets (t - A * mpr) }
    SetLength(tmpVect, N + 1);
    MatrixVectorMultiply(svdA, 1, N, 1, M, False, mpr, 1, M, 1, tmpVect, 1, N, 0);
    SetLength(Brack, N + 1);
    for i := 1 to N do
      Brack[i] := VectB[i - 1] - tmpVect[i];
    SetLength(tmpVect, 0);

    {computing the rest}
    SetLength(tmp, M + 1, N + 1);
    SetLength(work, Max(M, N) + 1);
    MatrixMatrixMultiply(InnerPart, 1, M, 1, M, False, svdATr, 1, M, 1, N, False,
      1, tmp, 1, M, 1, N, 0, work);
    SetLength(tmp1, N + 1, N + 1);
    MatrixMatrixMultiply(tmp, 1, M, 1, N, False, Cdi, 1, N, 1, N, False, 1, tmp1,
      1, N, 1, N, 0, work);
    SetLength(tmpVect, N + 1);
    MatrixVectorMultiply(tmp1, 1, N, 1, N, False, Brack, 1, N, 1, tmpVect, 1, N, 0);
    SetLength(tmp, 0, 0);
    SetLength(tmp1, 0, 0);

    {writing the results}
    for i := 1 to M do
      VectX[i - 1] := mpr[i] + tmpVect[i];
    for i := 1 to M do
      VectStd[i - 1] := Sqrt(InnerPart[i, i]);
      

    {Computing RMS}
    SetLength(work, N + 1);
    SetLength(tmpVect, M + 1);
    for i := 1 to M do
       tmpVect[i] := VectX[i - 1];

    MatrixVectorMultiply(svdA, 1, N, 1, M, False, tmpVect, 1, M, 1, work, 1, N, 0);
    suma := 0;
    for i := 1 to N do
      suma := suma + Power(work[i] - VectB[i - 1], 2);
    RMS := Sqrt(suma / N);



  finally
    Finalize(tmp);
    Finalize(tmp1);
    Finalize(work);
    Finalize(tmpVect);
    Finalize(svdA);
    Finalize(svdATr);
    Finalize(Cm);
    Finalize(Cmi);
    Finalize(Cd);
    Finalize(Cdi);
    Finalize(mpr);
    Finalize(InnerPart);
    Finalize(Brack);

//    frmMain.StatusBar1.SimpleText := 'End of SVD.';
  //  Application.ProcessMessages;
  end;

end;

procedure PseudoInverse(const Matr: TReal2DArray; const tol:double;
   var MatrInv: TReal2DArray);  {computes pseudoinverse of matrix Matr, tol is
   the lowest singular number allowed}
var
  i, j: integer;
  M, N: integer;
  W: TReal1DArray;
  U: TReal2DArray;
  VT: TReal2DArray;
  WMat: TReal2DArray;
  tmp: TReal2DArray;
  work: TReal1DArray;
begin
  M := Length(Matr) - 1;
  N := Length(Matr[0]) - 1;
  if M <> N then
    MessageDlg('Error: Pseudoinverse, M <> N.', mtError, [mbAbort], 0);

  try
    SetLength(WMat, N + 1, N + 1);

    {pseudoinverse of matrix}
    SVDDecomposition(Matr, M, M, 2, 2, 2, W, U, VT);

    {removing too small singular values}
    for i := 0 to M do
      for j := 0 to M do
        WMat[i, j] := 0;
    for i := 1 to M do
      begin
        if W[i] < tol then
          WMat[i, i] := 0
        else
          WMat[i, i] := 1 / W[i];
      end;
    for i := 1 to M do
      for j := 1 to M do
        if abs(U[i, j]) < NumZero then
          U[i, j] := 0;
    for i := 1 to M do
      for j := 1 to M do
        if abs(VT[i, j]) < NumZero then
          VT[i, j] := 0;

    SetLength(tmp, M + 1, M + 1);
    SetLength(work, M + 1);
    MatrixMatrixMultiply(U, 1, M, 1, M, False, WMat, 1, M, 1, M, False, 1, tmp,
      1, M, 1, M, 0, work);
    MatrixMatrixMultiply(tmp, 1, M, 1, M, False, VT, 1, M, 1, M, False, 1, MatrInv,
      1, M, 1, M, 0, work);

  finally
    Finalize(tmp);
    Finalize(work);
    Finalize(WMat);
    Finalize(U);
    Finalize(VT);
    Finalize(W);
  end;

end;

procedure Zapis();
var
  fw: TextFile;
  i, j: integer;
begin
  AssignFile(fw, 'vp.csv');
  Rewrite(fw);
  for i := 0 to (NrX - 1) do
    begin
      for j := 0 to (NrY - 1) do
        begin
          write(fw, FloatToStr(Rychlosti[0,i,j]));
          write(fw, #9);
        end;
      writeln(fw);
    end;
  writeln(fw, 'stop', #9, 'stop','stop', #9, 'stop');
  writeln(fw, 'stop', #9, 'stop','stop', #9, 'stop');
  for i := 0 to (NrX - 1) do
    begin
      for j := 0 to (NrY - 1) do
        begin
          write(fw, FloatToStr(VpZname[i,j]));
          write(fw, #9);
        end;
      writeln(fw);
    end;
  CloseFile(fw);
end;

procedure ZapisMatA(const NrOdp: integer; const ZacatekOdp: integer;
  const PocetGeofonu: integer);
var
  i, j: integer;
  fw: TextFile;
  ZacatekCtver: integer;
//  ZacatekOdp: integer;
//  a, b: integer;
begin
  AssignFile(fw, 'MatA.csv');
  Rewrite(fw);
//  a := Length(MatA[0]);
//  b := Length(MatA);
  ZacatekCtver := NrOdp * (PocetVrstev - 1) + PocetGeofonu * (PocetVrstev - 1);
//  ZacatekOdp := Length(TimeData);
  writeln(fw, IntToStr(Length(MatA)), #9, IntToSTr(Length(MatA[0])), #9,
    IntToStr(ZacatekCtver), #9, IntToStr(ZacatekOdp));
  for i := 0 to (Length(MatA[0]) - 1) do
    begin
      for j := 0 to (Length(MatA) - 1) do
        write(fw, FloatToStr(MatA[j, i]), #9);
//        write(fw, FloatToStrF(MatA[j, i], ffGeneral, 14, 14), #9);
      writeln(fw);
    end;
  CloseFile(fw);
end;

procedure ZapisVectB();
var
  i: integer;
  fw: TextFile;
begin
  AssignFile(fw, 'VectB.csv');
  Rewrite(fw);
  writeln(fw, IntToStr(Length(VectB)));
  for i := 0 to (Length(VectB) - 1) do
    writeln(fw, FloatToStr(VectB[i]));
//    writeln(fw, FloatToStrF(VectB[i], ffGeneral, 14, 14));
  CloseFile(fw);
end;

procedure PrectiVectX(const XSize: integer);
var
  i: integer;
  fr: TextFile;
begin
  AssignFile(fr, 'VectX.csv');
  Reset(fr);
  for i := 0 to (XSize - 1) do
    readln(fr, VectX[i]);
  CloseFile(fr);
end;

procedure PrectiVectStd(const XSize: integer);
var
  i: integer;
  tmp: double;
  fr: TextFile;
begin
  AssignFile(fr, 'VectX_std.csv');
  Reset(fr);
  for i := 0 to (XSize - 1) do
    readln(fr, tmp, VectStd[i]);
  CloseFile(fr);
end;

function ReadRMS(const RmsFileName: string): Double; {Funkce precte ze souboru hodnotu RMS}
var
  fr: TextFile;
begin
  AssignFile(fr, RmsFileName);
  Reset(fr);
  readln(fr, result);
  CloseFile(fr);
end;

{$IFDEF UNIX}

procedure SpustSVD();
var
  command: string;
begin
  command := './ComputeSVD.m MatA.csv VectB.csv ' + FloatToStr(RegClen);
  RunCommand(command);
end;

procedure RunCommand(Command: string);
var
//  Output: PIOFile;
  f: file;
begin
{  Output := libc.popen(PChar(Command), 'r');
  libc.pclose(output);}
//  Execlp(Command, envp);
  Assign(f, '/dev/null');
  popen(f, Command, 'r');
  pclose(f);
end;

{$ENDIF}


procedure ZapisVysledek(const fw_name: string);
var
  fw: TextFile;
  i, j, k: integer;
begin
  AssignFile(fw, fw_name);
  Rewrite(fw);
  write(fw, 'x', #9, 'y', #9);
  for j := 0 to (PocetVrstev - 1) do
    begin
      write(fw, 'v', IntToStr(j), #9);
      if j > 0 then write(fw, 'std_v', IntToStr(j), #9);
    end;

  for j := 0 to (PocetVrstev - 2) do
    begin
      write(fw, 'd', IntToStr(j), #9);
      write(fw, 'std_d', IntToStr(j), #9);
    end;
  writeln(fw);
  for i := 0 to (nrX - 1) do
    for k := 0 to (nrY - 1) do
      begin
        write(fw, FloatToStr(i*Scale + MinX), #9, FloatToStr(k * Scale + MinY), #9);
        for j := 0 to (PocetVrstev - 1) do
          begin
            write(fw, FloatToStr(Rychlosti[j, i, k]), #9);
{            if j > 0 then write(fw, FloatToStr(RychlostiStd[j, i, k] /
              power((1 / Rychlosti[j, i, k]), 2)), #9);}
            if j > 0 then write(fw, FloatToStr(RychlostiStd[j, i, k]), #9);
          end;
        for j := 0 to (PocetVrstev - 2) do
          begin
            if Mocnosti[j, i, k] <> -1 then
              begin
                write(fw, FloatToStr(Mocnosti[j, i, k] * Scale), #9);
                write(fw, FloatToStr(MocnostiStd[j, i, k] * Scale), #9);
              end
            else
              begin
                write(fw, FloatToStr(Mocnosti[j, i, k]), #9);
                write(fw, FloatToStr(MocnostiStd[j, i, k]), #9);
              end;
          end;
        writeln(fw);
      end;
  CloseFile(fw);
end;

procedure NajdiStejne(var OdpGf: TOdpGf; const ShotCoord, GeophCoord: TCoord;
  Identif: T2DDoubleArray; const TimeData: TTimeData);
var
  i, j, tmpId: integer;
  pocet: integer;
  pocetGf: integer; {celkovy pocet geofonu (bez duplicit) -- napr. 24}
  Gident, Iident: TDoubleArray;
  TmpGeophCoord: array of array[0..3] of double;
begin
  SetLength(Gident, Length(Identif));
  SetLength(Iident, Length(Identif));
  for i := 0 to (Length(Identif) - 1) do
    gident[i] := i; {vytvorime pole s cisly geofonu}
  for i := 0 to (Length(Identif) - 1) do
    Iident[i] := Identif[i, 1]; {vytvorime kopii Identif}
  {seradime podle Iident}
  Sort(Iident, Gident);
  pocet := 0;
  pocetGf := 0;
  {do TmpGeophCoord dame souradnice geofonu, identifikator a cislo vrstvy (proc to?!)}
  TmpId := -1;
  for i := 0 to (Length(Iident) - 1) do
    if TmpId <> Iident[i] then
      begin
        TmpId := trunc(Iident[i]);
        Inc(pocetGf);
        SetLength(TmpGeophCoord, pocetGf);
        TmpGeophCoord[pocetGf - 1, 0] := GeophCoord[trunc(gident[i]), 0];
        TmpGeophCoord[pocetGf - 1, 1] := GeophCoord[trunc(gident[i]), 1];
        TmpGeophCoord[pocetGf - 1, 2] := Iident[i];
        TmpGeophCoord[pocetGf - 1, 3] := TimeData[trunc(gident[i]), 2];
      end;

  {v OdpGf bude cislo zdroje, identifikator geofonu, cislo vrstvy}
  for i := 0 to (Length(TmpGeophCoord) - 1) do
    for j := 0 to (Length(ShotCoord) - 1) do
      begin
        if TmpGeophCoord[i, 0] = ShotCoord[j, 0] then
          if TmpGeophCoord[i, 1] = ShotCoord[j, 1] then
            begin
              Inc(pocet);
              SetLength(OdpGf, pocet);
              OdpGf[pocet - 1, 0] := j;
              OdpGf[pocet - 1, 1] := trunc(TmpGeophCoord[i, 2]);
              OdpGf[pocet - 1, 2] := trunc(TmpGeophCoord[i, 3]);
            end;
      end;
end;

procedure UrciProfilCoord(const ShotCoord, GeophCoord: TCoord;
  var ProfilCoordTransp: TDoubleArray);
var
  i, j: integer;
  pocet: integer;
  PomX: TDoubleArray;
  PocOdp, PocGf: integer;
begin
  pocet := Length(ShotCoord) + Length(GeophCoord);
  SetLength(PomX, pocet);
  PocOdp := Length(ShotCoord);
  PocGf := Length(GeophCoord);
  for i := 0 to (PocOdp - 1) do
    PomX[i] := ShotCoord[i, 0];
  for i := 0 to (PocGf - 1) do
    PomX[i + PocOdp] := GeophCoord[i, 0];
  Sort(PomX);
  SetLength(ProfilCoordTransp, 1);
  ProfilCoordTransp[0] := PomX[0];
  j := 0;
  for i := 1 to (Length(PomX) - 1) do
    begin
      if PomX[i] <> ProfilCoordTransp[j] then
        begin
          Inc(j);
          SetLength(ProfilCoordTransp, j + 1);
          ProfilCoordTransp[j] := PomX[i];
        end;
    end;
end;


Procedure Sort (var Key : TDoubleArray; var Data : TDoubleArray);

  Procedure QuickSort (L, R : Integer);
  var I, J, M : Integer;
    Begin
      Repeat
        I := L;
        J := R;
        M := (L + R) shr 1;
        Repeat
          While Key [I] < Key [M] do
            Inc (I);
          While Key [J] > Key [M] do
            Dec (J);
          if I <= J then
            begin
              Swap (Key [I], Key [J]);
              Swap (Data [I], Data [J]);
              if M = I then
                M := J else
                if M = J then
                  M := I;
              Inc (I);
              Dec (J);
            end;
        Until I > J;
        if L < J then
          QuickSort (L, J);
        L := I;
      Until I >= R;
    End;

var I : Integer;
  Begin
    Assert (Length (Key) = Length (Data), 'Sort pair must be of equal length.');
    I := Length (Key);
    if I > 0 then
      QuickSort (0, I - 1);
  End;


Procedure Sort (var V : TDoubleArray);

  Procedure QuickSort (L, R : Integer);
  var I, J, M : Integer;
    Begin
      Repeat
        I := L;
        J := R;
        M := (L + R) shr 1;
        Repeat
          While V [I] < V [M] do
            Inc (I);
          While V [J] > V [M] do
            Dec (J);
          if I <= J then
            begin
              Swap (V [I], V [J]);
              if M = I then
                M := J else
                if M = J then
                  M := I;
              Inc (I);
              Dec (J);
            end;
        Until I > J;
        if L < J then
          QuickSort (L, J);
        L := I;
      Until I >= R;
    End;

var I : Integer;
  Begin
    I := Length (V);
    if I > 0 then
      QuickSort (0, I - 1);
  End;

Procedure Swap (var X, Y : double);
var F : Extended;
  Begin
    F := X;
    X := Y;
    Y := F;
  End;


function IsNumeric(c: char): Boolean;
{copy&paste from IdGlobal}
begin
  // Do not use IsCharAlpha or IsCharAlphaNumeric - they are Win32 routines
  Result := Pos(c, '0123456789') > 0; {do not localize}
end;

function GetNrOfKnownVel(): Integer; {Funkce zjisti pocet znamych rychlosti
  zadanych v sgKnownVel}
var
  i: integer;
begin
  i := 1;
  while frmMain.sgKnownVel.Cells[1, i] <> '' do
    Inc(i);
  result := i - 1;
end;

function GetNrOfKnownDepths(): Integer; {Funkce zjisti pocet znamych hloubek
  zadanych v sgKnownDepths}
var
  i: integer;
begin
  i := 1;
  while frmMain.sgKnownDepths.Cells[1, i] <> '' do
    Inc(i);
  result := i - 1;
end;

function RefrMaxValue(const Data: array of Double): Double;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result < Data[I] then
      Result := Data[I];
end;

function RefrMinValue(const Data: array of Double): Double;
var
  I: Integer;
begin
  Result := Data[Low(Data)];
  for I := Low(Data) + 1 to High(Data) do
    if Result > Data[I] then
      Result := Data[I];
end;

end.

